/*
 * Synaptics Touchpad Controller
 * Implementation file.
 * By Daniel Meja V.
 */


#include "touchpad.h"
#include "ps2.h"

// Libraries included for debugging purposes, remove later
#include <avr/io.h>
#include "pin_access.h"



/** PROTOTYPES **/
void send_tp_arg(byte arg);
byte create_mode_byte(byte mode,  byte rate, byte sleep, byte gestures,
					byte packet_size, byte wmode);



/** PUBLIC FUNCTIONS **/
/*
 * Checks if the connected device is
 * a valid Synaptics touchpad
 */
int tp_check_valid_device() {
	int result;
	send_tp_arg(TPQUERY_IDTP);

	ps2_send(PS2COMMAND_STATUS_REQUEST);

	ps2_get(); // Minor not required

	result = ps2_get() != 0x47? -1 : 0;
	//SET(PS2ERROR_PORT, PS2ERR0R_PIN);
	ps2_get(); // Major not required

	return result;
}


/*
 * Assigns a value to the mode configuration byte.
 * mode 0 : Relative, 1 : Absolute.
 * rate 0 : 40packets/sec, 1 : 80packets/sec
 * sleep 0 : Sleep mode disabled, 1 : Enabled
 * gestures 0 : Tap and drag disabled, 1 : Enabled
 * packet_size 0 : Six-byte absolute packets, 1 : Eight-byte absolute packets
 * wmode 0 : Disable Enable Enhanced Absolute Packets with W (Finger Width) Information, 1 : Enabled
 */
void tp_set_mode_byte(byte mode,  byte rate, byte sleep, byte gestures,
					byte packet_size, byte wmode) {

	// Fixed sequence of sp/2 commands to set the touchpad mode byte

	send_tp_arg(create_mode_byte(mode, rate, sleep, gestures, packet_size, wmode));
	ps2_send(PS2COMMAND_SET_SAMPLE_RATE);
	ps2_send(0x14);

}

/*
 * Queries the touchpad for specific
 * device information.
 * query_type: Use one of the TPQUERY macros
 * Assigns res1 and res2 with the return values of the query.
 * Returns 0 if the device is  a valid Synaptics touchpad,
 * and -1 otherwise.
 */
int tp_query(byte query_type, byte *res1, byte *res2) {
	// Fixed sequence of sp/2 commands to query the touchpad
	int result = 0;
	send_tp_arg(query_type);
	ps2_send(PS2COMMAND_STATUS_REQUEST);
	*res1 = ps2_get();
	if (ps2_get() != 0x47)
		result = -1;
	*res2 = ps2_get();
	return result;
}

/*
 * Disables the data transmission
 * from the touchpad
 */
void tp_inhibit() {
	ps2_inhibit_device();
}

/*
 * Enables the data transmission
 * from the touchpad
 */
void tp_disinhibit() {
	ps2_disinhibit_device();
}

/*
 * Enables packet transmission
 */
void tp_enable_packet_transmission() {
	ps2_send(PS2COMMAND_ENABLE);
}

/*
 * Disables (stream mode) packet transmission
 */
void tp_disable_packet_transmission() {
	ps2_send(PS2COMMAND_DISABLE);
}

/*
 * Initializes the ports used to
 * access the touchpad
 */
void tp_init_ports() {
	ps2_init_ports();
}

/*
 * Receives a byte of data.
 * NOTE: It does NOT wait for a the clock pin to go low.
 * The pin should go low before calling this function.
 */
byte tp_receive_data() {
	return ps2_receive();
}

/*
 * Receives a byte of data.
 * NOTE: It DOES wait for a the clock pin to go low.
 */
byte tp_get_data() {
	return ps2_get();
}

/*
 * Parses an absolute packet of data. (Packet structure with Wmode = 1)
 * Receives a pointer to a byte array that has the 6 bytes of the packet.
 * Returns the parsed absolute packet
 */
tp_abs_packet_t tp_parse_absolute_packet (byte *packet) {
	tp_abs_packet_t abs_packet;

	byte *x_pos = (byte *)&(abs_packet.xpos);
	byte *y_pos = (byte *)&(abs_packet.ypos);

	//memcpy(x_pos, packet+4, 1);
	*(x_pos) = *(packet+4);
	*(x_pos + 1) = ((*(packet + 1))&0x0F) | ((*(packet + 3))&0x10);

	//memcpy(y_pos, packet+5, 1);
	*(y_pos) = *(packet+5);
	*(y_pos + 1) = ((*(packet + 1)>>4)&0x0F) | ((*(packet + 3)>>1)&0x10);

	abs_packet.zpress = *(packet + 2);

	abs_packet.wval = ((*(packet + 3)>>2)&0x01) | ((*(packet)>>1)&0x02) | ((*(packet)>>2)&0x0C);
	return abs_packet;

}


/*
 * Parses an absolute packet of data. (Packet structure with Wmode = 1)
 * Receives a pointer to a byte array that has the 6 bytes of the packet.
 * Assigns xpos, ypos, zpress, and wval with the appropriate parsed values.
 */

/*
void tp_parse_absolute_packet (byte *packet, unsigned int *xpos, unsigned int *ypos,
							byte *zpress, byte *wval) {
	byte *x_pos = (byte *)xpos;
	byte *y_pos = (byte *)ypos;
	memcpy(x_pos, packet+4, 1);
	*(x_pos + 1) = (*(packet + 1))&0x0F;
	*(x_pos + 1) |= (*(packet + 3))&0x10;

	memcpy(y_pos, packet+5, 1);
	*(y_pos + 1) = (*(packet + 1)>>4)&0x0F;
	*(y_pos + 1) |= (*(packet + 3)>>1)&0x10;

	memcpy(zpress, packet + 2, 1);

	*wval = (*(packet + 3)>>2)&0x01 | (*(packet)>>1)&0x02 | (*(packet)>>2)&0x0C;
}*/


/** PRIVATE FUNCTIONS **/

/*
 * Sends an argument for a touchpad command
 * that is embedded in sp/2 commands
 */
void send_tp_arg(byte arg) {
	byte i;
	for (i = 0; i < 4; i++) {
		ps2_send(PS2COMMAND_SET_RESOLUTION);
		ps2_send((arg >> (6 - i*2)) & 3); // Send the byte by groups of 2 bits
	}
}

/*
 * Creates a mode configuration byte with
 * the specified parameters.
 */
byte create_mode_byte(byte mode,  byte rate, byte sleep, byte gestures,
					byte packet_size, byte wmode) {

	return (mode<<7) | (rate<<6) | (sleep<<3) | (gestures<<2) | (packet_size<<1) | wmode;
}



