/*
 * Gestures Security System interface
 * By Daniel Meja V.
 */

/*
 * PINS:
 * D2: PS/2 Clock (red wire)
 * D3: PS/2 Data (brown wire)
 * A0: Action button (Start/Stop password input)
 * A1: Admin button
 * A6: Admit (green) LED
 * A7: Reject (red) LED
 *
 */

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <stdio.h>
#include <util/delay.h>
#include <string.h>
#include "pin_access.h"
#include "uart.h"
#include "touchpad.h"
#include "gestures.h"

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

// MACROS
#define ABSOLUTE_BYTE_COUNT 6
#define TPBUFFER_SIZE 20 // Needs to be bigger than ABSOLUTE_BYTE_COUNT
#define TPSCAN_TIME 50 // Check the touchpad every __ms
#define LEDBLINK_TIME 250
#define BUTTONCHECK_TIME 50 // Time to check the action button
#define NUM_LED_BLINKS 7

#define ACTION_BUTTON_PORT PINA
#define ACTION_BUTTON_PORTDIR DDRA
#define ACTION_BUTTON_PIN PINA0
#define ADMIN_BUTTON_PORT PINA
#define ADMIN_BUTTON_PORTDIR DDRA
#define ADMIN_BUTTON_PIN PINA1
#define ADMIT_LED_PORT PORTA
#define ADMIT_LED_PORTDIR DDRA
#define ADMIT_LED_PIN PINA6
#define REJECT_LED_PORT PORTA
#define REJECT_LED_PORTDIR DDRA
#define REJECT_LED_PIN PINA7

#define VALID_EEPROM_DATA 0x58 // Byte put into the eeprom to recognize that there is valid data
#define EEPROM_PW_ADDRESS 0x00
#define EEPROM_STATS_ADDRESS 0x400

#define PW_MAX_SIZE 150

// PC message types
#define PCMSG_SEND_PW 'P'
#define PCMSG_REQUEST_STATS 'S'
#define PCMSC_CLEAR_STATS 'C'
#define PCMSC_ACK 'A'

// TYPE DEFINITIONS
typedef enum {
	INTERFACE_IDLE,
	INTERFACE_RECEIVING_PW,
	INTERFACE_PW_ACCEPTED,
	INTERFACE_PW_DENIED,
	INTERFACE_RECEIVING_PC_PW,
	INTERFACE_SENDING_STATS,
	INTERFACE_CLEARING_STATS
} interface_state_t;

typedef enum {
	BUTTON_IDLE,
	BUTTON_MAYBE_PRESSED,
	BUTTON_PRESSED,
	BUTTON_MAYBE_RELEASED,
	BUTTON_RELEASED
} button_state_t;


// PROTOTYPES
void enable_tp_ISR();
void disable_tp_ISR();
void print_pw(gesture_raw_unit_t *apw, int apw_size);

// GLOBAL VARIABLES
volatile byte tpdata[TPBUFFER_SIZE], newdata;
volatile unsigned int tpnumbytes, time_scan, time_button, time_led;
byte blink_count;
byte button_actioned, action_received;
byte admin_button_actioned, admin_action_received;
volatile interface_state_t istate;
button_state_t bt_state, admin_bt_state;
volatile unsigned int pw_size;
volatile gesture_raw_unit_t pw[PW_MAX_SIZE]; // TODO: Consider making it volatile
volatile unsigned int pc_pw_size; // Size of the (model) password sent by the pc
volatile byte pc_pw_x; // Stores the x value (x and y are sent in different data bytes)
					   // to add it later to the pw.
volatile byte x_received; // Shows that the x value for the current pos (from pc)
						  // was already received.
volatile byte size_received; // Shows that the pc pw size was already received
volatile byte pc_pw_ready; // Done receiving pw from pc

byte reject_count, admit_count;

byte timeout;
byte admin_mode;

unsigned int model_pw_size;
gesture_raw_unit_t model_pw[PW_MAX_SIZE]; // The actual current password of the system


/*
 * External interrupt ISR
 */
ISR(INT0_vect) {
	tpdata[tpnumbytes++] = tp_receive_data();
	//tp_inhibit(); // Already inhibited in tp_get_data
	if (tpnumbytes < ABSOLUTE_BYTE_COUNT)
		tp_disinhibit();
	else
		newdata = 1;

	SET(EIFR, INTF0); // Clears the flag (for interrupts that happened while in the ISR)
}

/*
 * Timer 0 Compare Match ISR
 */
ISR(TIMER0_COMPA_vect) {
	if (time_scan < TPSCAN_TIME) {
		time_scan++;
	}
	if (time_button < BUTTONCHECK_TIME) {
		time_button++;
	}
	if(time_led < LEDBLINK_TIME) {
		time_led++;
	}
}

/*
 * USART Rx ISR
 */
ISR (USART0_RX_vect) {
	byte rx_value = UDR0;
	gesture_raw_unit_t pw_unit;
	// Only receives pc pw when idle
	if (istate == INTERFACE_IDLE || istate == INTERFACE_RECEIVING_PC_PW) {
		if (istate == INTERFACE_RECEIVING_PC_PW && pc_pw_ready == 0) {
			if (size_received == 0) {
				SET(ADMIT_LED_PORT, ADMIT_LED_PIN);
				size_received = 1;
				pc_pw_size = rx_value;
			} else if (x_received == 0) {
				pc_pw_x = rx_value;
				x_received = 1;
			} else { // Complete unit received, store in pw
				pw_unit.xpos = pc_pw_x;
				pw_unit.ypos = rx_value;
				// Z press ignored
				pw_unit.zpress = pw_unit.xpos == 0 && pw_unit.ypos == 0 ? 0 : 255;
				x_received = 0;
				pw[pw_size++] = pw_unit;
				if (pw_size == pc_pw_size) { // Done receiving pc pw
					pc_pw_ready = 1; // Signal for the main logic to continue
					size_received = 0;
					CLR(ADMIT_LED_PORT, ADMIT_LED_PIN);
				}
			}
		} else if (rx_value == PCMSG_SEND_PW) {
			istate = INTERFACE_RECEIVING_PC_PW;
			pw_size = 0;
		} else if (rx_value == PCMSG_REQUEST_STATS) {
			istate = INTERFACE_SENDING_STATS;
		} else if (rx_value == PCMSC_CLEAR_STATS) {
			istate = INTERFACE_CLEARING_STATS;
		}
	}
}

/*
 * Initializes the MCU and all the global variables
 */
void initialize() {
	// Rx Tx Debugging pins
	CLR(DDRD, PIND0);
	SET(DDRD, PIND1);

	// LED as outputs (initially off)
	SET(ADMIT_LED_PORTDIR, ADMIT_LED_PIN);
	SET(REJECT_LED_PORTDIR, REJECT_LED_PIN);
	CLR(ADMIT_LED_PORT, ADMIT_LED_PIN);
	CLR(REJECT_LED_PORT, REJECT_LED_PIN);

	// Buttons as inputs
	CLR(ACTION_BUTTON_PORTDIR, ACTION_BUTTON_PIN);
	CLR(ADMIN_BUTTON_PORTDIR, ADMIN_BUTTON_PIN);

    // Turn on pull-up resistors
    SET(PORTA, ACTION_BUTTON_PIN);
    SET(PORTA, ADMIN_BUTTON_PIN);

	// LED test pins
	DDRB = 0xFF;
	PORTB = 0xFF;


	// Set timer 0 for 1ms base
    TIMSK0= (1<<OCIE0A); // Turn on Compare Match ISR
    OCR0A = 249; // Compare register to 250 time ticks
    TCCR0B= 3; // Prescalar of 64
    TCCR0A= (1<<WGM01); // Clear on match

	newdata = 0;
	time_button = 0;
	time_led = 0;
	time_scan = 0;
	button_actioned = action_received = admin_button_actioned = admin_action_received = 0;
	pw_size = 0;
	pc_pw_size = 0;
	pc_pw_ready = 0;
	x_received = 0;
	size_received = 0;
	model_pw_size = 0;
	timeout = 0;
	admin_mode = 0;
	blink_count = 0;
	istate = INTERFACE_IDLE;
	bt_state = BUTTON_IDLE;
	admin_bt_state = BUTTON_IDLE;
	uart_init();
	stdout = stdin = stderr = &uart_str;
	UCSR0B |= (1<<RXCIE0); // Enable UART Rx ISR
	sei();
	fprintf(stdout,"Hello\n");
	// Check for password in EEPROM
	if (eeprom_read_byte(EEPROM_PW_ADDRESS) == VALID_EEPROM_DATA) {
		fprintf(stdout, "Valid data in EEPROM\n");
		model_pw_size = eeprom_read_word((uint16_t *)(EEPROM_PW_ADDRESS+1));
		fprintf(stdout, "Model pw size: %d\n", model_pw_size);
		eeprom_read_block(model_pw, (const void *)(EEPROM_PW_ADDRESS+3), model_pw_size*sizeof(gesture_raw_unit_t));
	}
	// Check for stats in EEPROM
	if (eeprom_read_word((uint16_t *)EEPROM_STATS_ADDRESS) == VALID_EEPROM_DATA) {
		fprintf(stdout, "Found stats\n");
		admit_count= eeprom_read_word((uint16_t *)(EEPROM_STATS_ADDRESS+1));
		reject_count= eeprom_read_word((uint16_t *)(EEPROM_STATS_ADDRESS+3));
		fprintf(stdout, "Admin count: %d, Reject count: %d\n", admit_count, reject_count);
	} else {
		admit_count = reject_count = 0;
	}

}

/*
 * Checks if the connected device is a valid
 * touchpad, and returns -1 if it's not.
 * Otherwise, initializes the touchpad in absolute mode,
 * and enables ISR for data reception.
 */
int init_touchpad() {
	int result;
	fprintf(stdout, "\n\n\n\n\n\n\n\n");
	tp_init_ports();
	tp_disinhibit();
	if ((result = tp_check_valid_device()) == -1) {
		fprintf(stdout, "Not a valid touchpad\nResult was %x\n", result&0xFF);
		return -1;
	}
	fprintf(stdout, "Valid touchpad!\n");
	tp_disable_packet_transmission();
	tp_set_mode_byte(1, 0, 0, 0, 0, 1); // Absolute mode, with "W" enabled
	tp_enable_packet_transmission();
	//tp_inhibit();
	enable_tp_ISR();
	return 0;
}

/*
 * Enables the external interrupt ISR for
 * data reception from the touchpad.
 */
void enable_tp_ISR() {
	CLR(EIMSK, INT0); // So that the change in EICRA does not trigger an interrupt
	EICRA = 2; // Falling Edge
	SET(EIFR, INTF0); // Clears the flag
	SET(EIMSK, INT0); // Enable external interrupt in PIN D2

}

/*
 * Disables the external interrupt ISR for
 * data reception from the touchpad.
 */
void disable_tp_ISR() {
	CLR(EIMSK, INT0);
	EICRA = 0;
}

/*
 * Initializes the reception of data
 * from the touchpad
 */
void init_tp_rx() {
	newdata = 0;
	tp_disinhibit();
	tpnumbytes = 0;
}

/*
 * Scans and returns the data obtained from the touchpad.
 * If no new data, the returned gesture_data_t variable
 * has all fields as zero.
 */
gesture_raw_unit_t scan_tp() {
	tp_abs_packet_t packet;
	gesture_raw_unit_t gdata;
	if (newdata) {
		packet = tp_parse_absolute_packet((byte *)tpdata);
		return touchpad2gesture_raw_unit(packet);
	}
	gdata.xpos = 0;
	gdata.ypos = 0;
	gdata.zpress = 0;
	return gdata;
}

/*
 * Initialization for receiving a new password
 */
void init_pw_rx() {
	pw_size = 0;
}

/*
 * Updates the password with the received data.
 * Returns -1 if the size limit was exceeded.
 */
int update_pw(gesture_raw_unit_t data) {
	if (pw_size == PW_MAX_SIZE - 1)
		return -1;
	pw[pw_size++] = data;
	return 0;
}
/*
 * DEBUG function that prints the pw
 */
void print_pw(gesture_raw_unit_t *apw, int apw_size) {
	int i;
	fprintf(stdout, "\n\nThe password is:\n\n\n");
	for (i = 0; i < apw_size; i++) {
		fprintf(stdout, "PW UNIT %d:\nx:%d\ny:%d\nz:%d\n\n",
				i, apw[i].xpos, apw[i].ypos, apw[i].zpress);
	}
	fprintf(stdout, "\n\n\n");
}

/*
 * Sets the password and saves it to EEPROM
 */
void save_pw() {
	model_pw_size = pw_size;
	memcpy(model_pw, (void *)pw, model_pw_size*sizeof(gesture_raw_unit_t));
	// Save in EEPROM
	eeprom_write_word((uint16_t *)(EEPROM_PW_ADDRESS+1), model_pw_size);
	eeprom_write_block(model_pw,(void *)(EEPROM_PW_ADDRESS+3),  model_pw_size*sizeof(gesture_raw_unit_t));
	eeprom_write_byte(EEPROM_PW_ADDRESS, VALID_EEPROM_DATA); // So that valid data is recognized
}

/*
 * Checks if the entered password
 * is valid
 */
int valid_pw() {
	return gesture_check_match(model_pw, model_pw_size, (gesture_raw_unit_t *)pw, pw_size);
}

/*
 * Sets the admit count and reject count stats to EEPROM
 */
void save_stats() {
	eeprom_write_word((uint16_t *)(EEPROM_STATS_ADDRESS+1), admit_count);
	eeprom_write_word((uint16_t *)(EEPROM_STATS_ADDRESS+3),  reject_count);
	eeprom_write_word((uint16_t *)EEPROM_STATS_ADDRESS, VALID_EEPROM_DATA); // So that valid data is recognized
}

/*
 * Sends the stats to the computer
 */
void send_stats() {
	fputc(PCMSC_ACK, stdout);
	fputc(admit_count, stdout);
	fputc(reject_count, stdout);
}

/*
 * Clear the stats and saves to EEPROM
 */
void clear_stats() {
	admit_count = reject_count = 0;
	save_stats();
}

/*
 * FSM that checks if the action button was
 * pressed. If it was, then button_actioned is set to 1.
 * Once the action is detected, the FSM waits until
 * action_received is 1 to go back to checking the button.
 */
void check_button(byte pin_read, button_state_t *thebt_state,
				  byte *abutton_actioned, byte *theaction_received) {
	switch(*thebt_state) {
	case BUTTON_IDLE:
		if(pin_read == 0) {
			*thebt_state = BUTTON_MAYBE_PRESSED;
		}
		break;
	case BUTTON_MAYBE_PRESSED:
		if(pin_read == 0) {
			*thebt_state = BUTTON_PRESSED;
			*abutton_actioned = 1;
		} else
			*thebt_state = BUTTON_IDLE;
		break;
	case BUTTON_PRESSED:
		if(pin_read == 1) {
			*thebt_state = BUTTON_MAYBE_RELEASED;
		}
		break;
	case BUTTON_MAYBE_RELEASED:
		if(pin_read == 1) {
			*thebt_state = BUTTON_RELEASED;
		} else {
			*thebt_state = BUTTON_PRESSED;
		}
		break;
	case BUTTON_RELEASED:
		if (*theaction_received) {
			*theaction_received = 0;
			*thebt_state = BUTTON_IDLE;
		}
		break;
	}
}

/*
 * Updates the main interface FSM.
 */
void update_interface_state() {
	switch (istate) {
	case INTERFACE_IDLE:
		if (button_actioned) {
			button_actioned = 0;
			action_received = 1;
			fprintf(stdout, "Enter password\n");
			istate = INTERFACE_RECEIVING_PW;
			SET(ADMIT_LED_PORT, ADMIT_LED_PIN);
			init_tp_rx();
			init_pw_rx();
		} else if (admin_button_actioned) {
			admin_button_actioned = 0;
			admin_action_received = 1;
			admin_mode = 1;
			istate = INTERFACE_IDLE;
			fprintf(stdout, "Entered admin mode\n");
		}
		break;
	case INTERFACE_RECEIVING_PW:
		if (timeout) {
			timeout = 0;
			istate = INTERFACE_PW_DENIED;
			CLR(ADMIT_LED_PORT, ADMIT_LED_PIN);
		}
		else if (button_actioned) {
			button_actioned = 0;
			action_received = 1;
			CLR(ADMIT_LED_PORT, ADMIT_LED_PIN);
			//tp_inhibit();  // Would cause an interrupt and get stuck.
							 // The touchpad is inhibited in the interrupt anyway.

			if (admin_mode) {
				admin_mode = 0;
				fprintf(stdout, "Password set by admin\nOut of admin mode..\n\n");
				save_pw();
				istate = INTERFACE_IDLE;
			} else if (valid_pw())
				istate = INTERFACE_PW_ACCEPTED;
			else
				istate = INTERFACE_PW_DENIED;

		}
		break;
	case INTERFACE_PW_ACCEPTED:
		if (blink_count == 0) {
			fprintf(stdout, "Password Accepted\n\n");
			admit_count++;
			save_stats();
		}
		if (time_led >= LEDBLINK_TIME) {
			time_led = 0;
			FLIP(ADMIT_LED_PORT, ADMIT_LED_PIN);
			blink_count++;
		}
		if (blink_count == NUM_LED_BLINKS) {
			blink_count = 0;
			CLR(ADMIT_LED_PORT, ADMIT_LED_PIN);

			istate = INTERFACE_IDLE;
		}
		break;
	case INTERFACE_PW_DENIED:
		if (blink_count == 0) {
			fprintf(stdout, "Invalid Password\n\n");
			reject_count++;
			save_stats();
		}
		if (time_led >= LEDBLINK_TIME) {
			time_led = 0;
			FLIP(REJECT_LED_PORT, REJECT_LED_PIN);
			blink_count++;
		}
		if (blink_count == NUM_LED_BLINKS) {
			blink_count = 0;
			CLR(REJECT_LED_PORT, REJECT_LED_PIN);
			istate = INTERFACE_IDLE;
		}
		break;
	case INTERFACE_RECEIVING_PC_PW:
		if (pc_pw_ready) {
			pc_pw_ready = 0;
			fprintf(stdout, "PasSword set from pc\n\n");
			save_pw();
			istate = INTERFACE_PW_ACCEPTED;
		}
		break;
	case INTERFACE_SENDING_STATS:
		send_stats();
		istate = INTERFACE_IDLE;
		break;
	case INTERFACE_CLEARING_STATS:
		clear_stats();
		istate = INTERFACE_IDLE;
		break;
	default:
		break;
	}
}

int main(void) {
	initialize();
	if (init_touchpad() == -1) // Invalid touchpad
		return -1;
	while (1) {
		if (time_button >= BUTTONCHECK_TIME){
			check_button(READ(ACTION_BUTTON_PORT, ACTION_BUTTON_PIN), &bt_state,
						 &button_actioned, &action_received);
			check_button(READ(ADMIN_BUTTON_PORT, ADMIN_BUTTON_PIN), &admin_bt_state,
						 &admin_button_actioned, &admin_action_received);
			time_button = 0;
		}

		update_interface_state();

		if (time_scan >= TPSCAN_TIME) {
			if (istate == INTERFACE_RECEIVING_PW) {
				if (update_pw(scan_tp()) == -1) {
					// Size limit reached
					timeout = 1;
				} else
					init_tp_rx(); // Re-enable reception from the touchpad
			}
			time_scan = 0;
		}
	}
}


