/*
 * Synaptics Touchpad Controller
 * Interface file.
 * By Daniel Meja V.
 */

#ifndef __TOUCHPAD_CONTROLLER_H__

#define __TOUCHPAD_CONTROLLER_H__

#include "ps2.h"

#define TPQUERY_IDTP 0x00
#define TPQUERY_READ_TP_MODES 0x01
#define TP_READCAPABILITIES 0x02
#define TP_READ_MODEL_ID 0x03
#define TP_READ_SERIAL_NUMBER_PREFIX 0x06
#define TP_READ_SERIAL_NUMBER_SUFIX 0x07
#define TP_READ_RESOLUTIONS 0x08

#define TP_LOWER_BEZEL_XPOS 1472
#define TP_HIGHER_BEZEL_XPOS 5472
#define TP_LOWER_BEZEL_YPOS 1408
#define TP_HIGHER_BEZEL_YPOS 4448


typedef struct {
	unsigned int xpos;
	unsigned int ypos;
	byte zpress;
	byte wval;
} tp_abs_packet_t;

/*
 * Checks if the connected device is
 * a valid Synaptics touchpad
 */
int tp_check_valid_device();


/*
 * Assigns a value to the mode configuration byte.
 * mode 0 : Relative, 1 : Absolute.
 * rate 0 : 40packets/sec, 1 : 80packets/sec
 * sleep 0 : Sleep mode disabled, 1 : Enabled
 * gestures 0 : Tap and drag disabled, 1 : Enabled
 * packet_size 0 : Six-byte absolute packets, 1 : Eight-byte absolute packets
 * wmode 0 : Disable Enable Enhanced Absolute Packets with W (Finger Width) Information, 1 : Enabled
 */
void tp_set_mode_byte(byte mode,  byte rate, byte sleep, byte gestures,
					byte packet_size, byte wmode);

/*
 * Queries the touchpad for specific
 * device information.
 * query_type: Use one of the TPQUERY macros
 * Assigns minor and major with the minor and major versions
 * of the touchpad.
 * Returns 0 if the device is  a valid Synaptics touchpad,
 * and -1 otherwise.
 */
int tp_query(byte query_type, byte *minor, byte *major);

/*
 * Enables (stream mode) packet transmission
 */
void tp_enable_packet_transmission();

/*
 * Disables (stream mode) packet transmission
 */
void tp_disable_packet_transmission();


/*
 * Inhibits the data transmission from the touchpad
 * by pulling the clock line down
 */
void tp_inhibit();

/*
 * Disinhibits the data transmission from the touchpad
 * by releasing the clock line (letting it go up)
 */
void tp_disinhibit();

/*
 * Initializes the ports used to
 * access the touchpad
 */
void tp_init_ports();

/*
 * Receives a byte of data.
 * NOTE: It does NOT wait for a the clock pin to go low.
 * This should happen before calling this function.
 */
byte tp_receive_data();

/*
 * Receives a byte of data.
 * NOTE: It DOES wait for a the clock pin to go low.
 */
byte tp_get_data();

/*
 * Parses an absolute packet of data. (Packet structure with Wmode = 1)
 * Receives a pointer to a byte array that has the 6 bytes of the packet.
 * Returns the parsed absolute packet
 */
tp_abs_packet_t tp_parse_absolute_packet (byte *packet);

#endif
